const { EmbedBuilder, ApplicationCommandOptionType } = require("discord.js");
const { MESSAGES, EMBED_COLORS } = require("@root/config");

/**
 * @type {import("@structures/Command")}
 */
module.exports = {
  name: "lyric",
  description: "get lyrics for the current song or a specified song",
  category: "MUSIC",
  botPermissions: ["EmbedLinks"],
  command: {
    enabled: true,
    usage: "[song name]",
  },
  slashCommand: {
    enabled: true,
    options: [
      {
        name: "query",
        description: "song name to search lyrics for",
        type: ApplicationCommandOptionType.String,
        required: false,
      },
    ],
  },

  async messageRun(message, args) {
    const query = args.join(" ");
    const response = await getLyric(message, query);
    await message.safeReply(response);
  },

  async interactionRun(interaction) {
    const query = interaction.options.getString("query");
    const response = await getLyric(interaction, query);
    await interaction.followUp(response);
  },
};

async function getLyric({ client, guild, member }, query) {
  const player = client.musicManager.getPlayer(guild.id);

  if (!player) {
    return "🚫 There's no active music player in this server.";
  }

  let track;
  let lyrics;

  try {
    if (!query) {
      if (!player.queue.current) {
        return "🚫 No music is currently playing!";
      }
      track = player.queue.current;
    } else {
      const result = await player.search({ query }, member.user);
      if (!result || result.loadType === "error" || result.loadType === "empty") {
        return "Failed to find any tracks for the query.";
      }
      track = result.tracks[0];
    }

    const node = player.node;
    const baseUrl = (node.options.port !== 80 && node.options.secure)
      ? `https://${node.options.host}:${node.options.port}`
      : `http://${node.options.host}:${node.options.port}`;

    // Try default lyrics endpoint first
    try {
      const defaultLyricsUrl = `${baseUrl}/v4/sessions/${node.sessionId}/players/${guild.id}/lyrics`;
      const defaultRes = await fetch(defaultLyricsUrl, {
        headers: { Authorization: node.options.authorization }
      });

      if (defaultRes.ok) {
        lyrics = await defaultRes.json();
        if (lyrics) {
          return createLyricsEmbed(lyrics, member, track);
        }
      }
    } catch (err) {
      client.logger.debug("Default lyrics endpoint failed:", err);
    }

    // Fallback to genius search if default endpoint fails
    const geniusUrl = `${baseUrl}/v4/lyrics/search?query=${encodeURIComponent(track.info.title + " " + track.info.author)}&source=genius`;
    const geniusRes = await fetch(geniusUrl, {
      headers: { Authorization: node.options.authorization }
    });

    if (!geniusRes.ok) {
      if (geniusRes.status === 404) return "No lyrics found for this song.";
      throw new Error(`Failed to fetch lyrics: ${geniusRes.status} ${geniusRes.statusText}`);
    }

    lyrics = await geniusRes.json();
    return createLyricsEmbed(lyrics, member, track);

  } catch (error) {
    client.logger.error("Lyric Command Error:", error);
    return "An error occurred while fetching the lyrics. Please try again later.";
  }
}

function createLyricsEmbed(lyrics, member, track) {
  if (!lyrics) {
    return "No lyrics found for this song.";
  }

  const embed = new EmbedBuilder()
    .setColor(EMBED_COLORS.BOT_EMBED)
    .setTitle(`${track.info.author} - ${track.info.title}`)
    .setThumbnail(track.info.artworkUrl)
    .setFooter({ text: `Requested by: ${member.user.displayName} | Source: ${lyrics.source || 'Unknown'}` });

  let ltext;
  if (lyrics.type === 'text') {
    ltext = lyrics.text;
  } else if (lyrics.type === 'timed') {
    ltext = lyrics.lines.map(line => line.line).join('\n');
  } else {
    return "Unsupported lyrics format";
  }

  embed.setDescription(ltext.length > 4096 ? ltext.slice(0, 4093) + "..." : ltext);

  return { embeds: [embed] };
}
