const { EmbedBuilder, ButtonBuilder, ButtonStyle, ActionRowBuilder } = require("discord.js");
const { EMBED_COLORS } = require("@root/config");

module.exports = async (client, player, track) => {
  const guild = client.guilds.cache.get(player.guildId);
  if (!guild) return;

  if (!player.textChannelId || !track) return;

  const channel = guild.channels.cache.get(player.textChannelId);
  if (!channel) return;

  if (player.get("autoplay") === true) {
    await player.queue.previous.push(track);
  }

  if (player.voiceChannelId) {
    await client.utils.setVoiceStatus(client, player.voiceChannelId, `Paying: **${track.info.title}**`);
  }

  const previous = await player.queue.shiftPrevious();

  const row1 = (player) =>
    new ActionRowBuilder().addComponents(
      new ButtonBuilder()
        .setCustomId("volumedown")
        .setEmoji("🔉")
        .setStyle(ButtonStyle.Secondary),
      new ButtonBuilder()
        .setCustomId("previous")
        .setEmoji("⏮️")
        .setStyle(ButtonStyle.Secondary)
        .setDisabled(!previous),
      new ButtonBuilder()
        .setCustomId("pause")
        .setEmoji(player.paused ? "▶️" : "⏸️")
        .setStyle(player.paused ? ButtonStyle.Success : ButtonStyle.Secondary),
      new ButtonBuilder()
        .setCustomId("skip")
        .setEmoji("⏭️")
        .setStyle(ButtonStyle.Secondary),
      new ButtonBuilder()
        .setCustomId("volumeup")
        .setEmoji("🔊")
        .setStyle(ButtonStyle.Secondary)
    );

  const row2 = (player) =>
    new ActionRowBuilder().addComponents(
      new ButtonBuilder()
        .setCustomId("autoplay")
        .setEmoji("♾️")
        .setStyle(player.get("autoplay") ? ButtonStyle.Success : ButtonStyle.Secondary),
      new ButtonBuilder()
        .setCustomId("loop")
        .setEmoji("🔁")
        .setStyle(player.repeatMode === "off" ? ButtonStyle.Secondary : ButtonStyle.Success),
      new ButtonBuilder()
        .setCustomId("stop")
        .setEmoji("⏹️")
        .setStyle(ButtonStyle.Danger),
      new ButtonBuilder()
        .setCustomId("shuffle")
        .setEmoji("🔀")
        .setStyle(ButtonStyle.Secondary),
      new ButtonBuilder()
        .setCustomId("queue")
        .setEmoji("📑")
        .setStyle(ButtonStyle.Secondary)
    );

  const msg = await channel.safeSend({
    embeds: [
      new EmbedBuilder()
        .setColor(EMBED_COLORS.BOT_EMBED)
        .setAuthor({ name: "Track Started" })
        .setDescription(`🎶 **Now playing [${track.info.title}](${track.info.uri})**`)
        .setThumbnail(track.info.artworkUrl)
        .setFooter({
          text: `Requested by: ${track.requester.displayName}`,
        })
        .addFields(
          {
            name: "Duration",
            value: track.info.isStream ? "Live" : client.utils.formatTime(track.info.duration),
            inline: true,
          },
          { name: "Author", value: track.info.author || "Unknown", inline: true }
        ),
    ],
    components: [row1(player), row2(player)],
  });

  if (msg) player.set("message", msg);

  const collector = msg.createMessageComponentCollector({
    filter: async (int) => {
      const sameVc = int.guild.members.me.voice.channelId === int.member.voice.channelId;
      return sameVc;
    },
  });

  collector.on("collect", async (int) => {
    if (!int.isButton()) return;

    await int.deferReply({ ephemeral: true });
    let description;

    switch (int.customId) {
      case "previous":
        description = previous ? "Playing the previous track..." : "No previous track available";
        if (previous) player.play({ clientTrack: previous });
        break;

      case "pause":
        if (player.paused) {
          player.resume();
          description = "Track resumed";
        } else {
          player.pause();
          description = "Track paused";
        }
        await msg.edit({ components: [row1(player), row2(player)] });
        break;

      case "stop":
        player.stopPlaying(true, false);
        description = "Playback stopped";
        break;

      case "skip":
        description = player.queue.tracks.length > 0 ? "Skipped to the next track" : "The queue is empty!";
        if (player.queue.tracks.length > 0) player.skip();
        break;

      case "shuffle":
        if (player.queue.tracks.length < 2) {
          description = "The queue is too short to shuffle!";
        } else {
          player.queue.shuffle();
          description = "The queue has been shuffled!";
        }
        break;

      case "volumeup":
        const newVolUp = Math.min(player.volume + 10, 100);
        await player.setVolume(newVolUp);
        description = `Volume increased to \`${newVolUp}%\``;
        break;

      case "volumedown":
        const newVolDown = Math.max(player.volume - 10, 0);
        await player.setVolume(newVolDown);
        description = `Volume decreased to \`${newVolDown}%\``;
        break;

      case "loop":
        const currentMode = player.repeatMode;
        if (currentMode === "off") {
          player.setRepeatMode("track");
          description = "Loop mode set to `track`";
        } else if (currentMode === "track") {
          player.setRepeatMode("queue");
          description = "Loop mode set to `queue`";
        } else {
          player.setRepeatMode("off");
          description = "Loop mode `disabled`";
        }
        await msg.edit({ components: [row1(player), row2(player)] });
        break;

      case "autoplay":
        const autoplay = !player.get("autoplay");
        player.set("autoplay", autoplay);
        description = autoplay ? "Autoplay activated!" : "Autoplay deactivated";
        await msg.edit({ components: [row1(player), row2(player)] });
        break;

      case "queue":
        const queue = player.queue;
        const tracks = queue.tracks.slice(0, 10); // Get first 10 tracks
        
        const queueEmbed = new EmbedBuilder()
          .setColor(EMBED_COLORS.BOT_EMBED)
          .setAuthor({ name: "Music Queue" })
          .setDescription(
            tracks.length > 0
              ? tracks
                  .map((track, i) => `${i + 1}. [${track.info.title}](${track.info.uri})`)
                  .join("\n")
              : "No tracks in queue"
          )
          .addFields({
            name: "Now Playing",
            value: `[${queue.current.info.title}](${queue.current.info.uri})`,
          });

        if (tracks.length > 0) {
          queueEmbed.setFooter({
            text: `${queue.tracks.length} track(s) in queue`,
          });
        }

        description = "Queue sent as new message";
        await channel.send({ embeds: [queueEmbed] });
        break;
    }

    await int.followUp({
      embeds: [new EmbedBuilder().setDescription(description).setColor(EMBED_COLORS.BOT_EMBED)],
    });
  });
};