const { COLORS } = require("@src/data.json");
const { readdirSync, lstatSync } = require("fs");
const { join, extname } = require("path");
const permissions = require("./permissions");

module.exports = class Utils {
  /**
 * {@link setTimeout} but returns function that clears timeout when called
 * @param {()=>void}fn Timeout function
 * @param {number} timeout Timeout in millis
 */
  static abortable(timeout, fn) {
    const id = setTimeout(fn, timeout)
    return () => clearTimeout(id)
  }
  /**
   * {@link AbortSignal.timeout} but can be aborted
   * @param {()=>void}fn Timeout function
   * @param {number} timeout Timeout in millis
   */
  static abortableSignal(timeout, fn) {
    const ctrl = new AbortController
    return Object.assign(
      Utils.abortable(timeout, () => ctrl.abort(fn?.())),
      { signal: ctrl.signal }
    )
  }

  /**
   * Checks if a string contains a URL
   * @param {string} text
   */
  static containsLink(text) {
    return /(https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|www\.[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9]+\.[^\s]{2,}|www\.[a-zA-Z0-9]+\.[^\s]{2,})/.test(
      text
    );
  }

  /**
   * Checks if a string is a valid discord invite
   * @param {string} text
   */
  static containsDiscordInvite(text) {
    return /(https?:\/\/)?(www.)?(discord.(gg|io|me|li|link|plus)|discorda?p?p?.com\/invite|invite.gg|dsc.gg|urlcord.cf)\/[^\s/]+?(?=\b)/.test(
      text
    );
  }

  /**
   * Returns a random number below a max
   * @param {number} max
   */
  static getRandomInt(max) {
    return Math.floor(Math.random() * max);
  }

  /**
   * Checks if a string is a valid Hex color
   * @param {string} text
   */
  static isHex(text) {
    return /^#[0-9A-F]{6}$/i.test(text);
  }

  /**
   * Checks if a string is a valid Hex color
   * @param {string} text
   */
  static isValidColor(text) {
    if (COLORS.indexOf(text) > -1) {
      return true;
    } else return false;
  }

  /**
   * Returns hour difference between two dates
   * @param {Date} dt2
   * @param {Date} dt1
   */
  static diffHours(dt2, dt1) {
    let diff = (dt2.getTime() - dt1.getTime()) / 1000;
    diff /= 60 * 60;
    return Math.abs(Math.round(diff));
  }

  /**
   * Returns remaining time in days, hours, minutes and seconds
   * @param {number} timeInSeconds
   */
  static timeformat(timeInSeconds) {
    const days = Math.floor((timeInSeconds % 31536000) / 86400);
    const hours = Math.floor((timeInSeconds % 86400) / 3600);
    const minutes = Math.floor((timeInSeconds % 3600) / 60);
    const seconds = Math.round(timeInSeconds % 60);
    return (
      (days > 0 ? `${days} days, ` : "") +
      (hours > 0 ? `${hours} hours, ` : "") +
      (minutes > 0 ? `${minutes} minutes, ` : "") +
      (seconds > 0 ? `${seconds} seconds` : "")
    );
  }

  /**
   * Converts duration to milliseconds
   * @param {string} duration
   */
  static durationToMillis(duration) {
    return (
      duration
        .split(":")
        .map(Number)
        .reduce((acc, curr) => curr + acc * 60) * 1000
    );
  }

  /**
   * Returns time remaining until provided date
   * @param {Date} timeUntil
   */
  static getRemainingTime(timeUntil) {
    const seconds = Math.abs((timeUntil - new Date()) / 1000);
    const time = Utils.timeformat(seconds);
    return time;
  }

  /**
   * @param {import("discord.js").PermissionResolvable[]} perms
   */
  static parsePermissions(perms) {
    const permissionWord = `permission${perms.length > 1 ? "s" : ""}`;
    return "`" + perms.map((perm) => permissions[perm]).join(", ") + "` " + permissionWord;
  }

  /**
   * Recursively searches for a file in a directory
   * @param {string} dir
   * @param {string[]} allowedExtensions
   */
  static recursiveReadDirSync(dir, allowedExtensions = [".js"]) {
    const filePaths = [];
    const readCommands = (dir) => {
      const files = readdirSync(join(process.cwd(), dir));
      files.forEach((file) => {
        const stat = lstatSync(join(process.cwd(), dir, file));
        if (stat.isDirectory()) {
          readCommands(join(dir, file));
        } else {
          const extension = extname(file);
          if (!allowedExtensions.includes(extension)) return;
          const filePath = join(process.cwd(), dir, file);
          filePaths.push(filePath);
        }
      });
    };
    readCommands(dir);
    return filePaths;
  }

  /**
   * Formats milliseconds into days, hours, minutes, and seconds
   * @param {number} ms - Milliseconds
   * @returns {string} - Formatted time string
   */
  static formatTime(ms) {
    return ms < 1000
      ? `${ms / 1000}s`
      : ["d", "h", "m", "s"]
          .map((unit, i) => {
            const value = [864e5, 36e5, 6e4, 1e3][i];
            const amount = Math.floor(ms / value);
            ms %= value;
            return amount ? `${amount}${unit}` : null;
          })
          .filter((x) => x !== null)
          .join(" ") || "0s";
  }

  /**
   * Parses a time string into milliseconds
   * @param {string} string - The time string (e.g., "1d", "2h", "3m", "4s")
   * @returns {number} - The time in milliseconds
   */
  static parseTime(string) {
    const time = string.match(/([0-9]+[dhms])/g);
    if (!time) return 0;
    return time.reduce((ms, t) => {
      const unit = t[t.length - 1];
      const amount = Number(t.slice(0, -1));
      return ms + amount * { d: 864e5, h: 36e5, m: 6e4, s: 1e3 }[unit];
    }, 0);
  }

  /**
   * Updates voice channel status
   * @param {string} channel - The voice channel ID
   * @param {string} status - The status to update
   * @param {object} client - The bot client
   */
  static async setVoiceStatus(client, channelId, message) {
    const url = `/channels/${channelId}/voice-status`;
    const payload = { status: message };
    await client.rest.put(url, { body: payload }).catch(() => {});
  }
};
